/* LINTLIBRARY */
#include <sys/param.h>
#include <sys/time.h>
#include <sys/vnode.h>
#include <sys/ucred.h>
#define KERNEL
#include <sys/file.h>
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>
#undef KERNEL
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/ioctl_compat.h>	/* XXX NTTYDISC is too well hidden */
#include <sys/tty.h>
#include <sys/conf.h>
#include <sys/rlist.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <err.h>
#include <kvm.h>
#include <limits.h>
#include <nlist.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/dkstat.h>
#ifdef HAVE_DEVSTAT
#include <devstat.h>
#endif
#include <unistd.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/time.h>
#include <sys/mount.h>
#include <nfs/rpcv2.h>
#include <nfs/nfsv2.h>
#include <nfs/nfs.h>
#include <osreldate.h>
#include <fcntl.h>

#if (__FreeBSD_version >= 199512)
#include <sys/vmmeter.h>
#endif
#if (__FreeBSD_version >= 300000)
#include <net/if_var.h>
#endif

extern int num_bars();
extern void bar_items(/* int nbars, int items[] */);
extern char **label_bars(/* int nbars */);
extern void init_bars(/* int nbars */);
extern void display_bars(/* int nbars */);
extern char *xmalloc(/* int nbytes */);
extern void shorten(/* char *hname */);
extern char *kernelSymbols;
void    swapmode __P((int*, int*));
void    swapmode_pstat __P((int*, int*));
int getkval __P((unsigned long, int *, int, char *));
extern void draw_bar(/*int bar_num, int *states, int num_states*/);
void scale_bar(int, int, int, int *, int);

extern int pcount;
extern int cpuflag, loadflag, memflag, swapflag, diskflag, pktflag, nfsflag;
long	cp_time[CPUSTATES];
long	cp_old[CPUSTATES];
#ifndef HAVE_DEVSTAT
long    disk_xfer[DK_NDRIVE], old_disk_xfer[DK_NDRIVE];
#endif
int	kmem;			/* file descriptor of /dev/kmem. */
kvm_t   *kd;
char    *nlistf = NULL;
char    *memf   = NULL;
off_t ifnetaddr;
struct ifnet ifnet;

struct packet {
  int in_out, collisions;
} packets, old_packets;

#ifdef HAVE_DEVSTAT
long generation;
devstat_select_mode select_mode;
struct devstat_match *matches;
int num_matches;
int num_selected, num_selections;
long select_generation;
static struct statinfo cur, last;
int num_devices;
struct device_selection *dev_select;
char nodisk;

void init_devstat(void);
int get_devstat(void);
#endif /* HAVE_DEVSTAT */

static struct nfsstats nfsstats;
static struct _nfsStats {
    int nfsServer, nfsClient;
} nfsStats, old_nfsStats;

struct nlist	nl[] = {
#define	X_CP_TIME	0
	{ "_cp_time" },
#define VM_SWAPLIST	1
	{ "_swaplist" },/* list of free swap areas */
#define VM_SWDEVT	2
	{ "_swdevt" },	/* list of swap devices and sizes */
#define VM_NSWAP	3
	{ "_nswap" },	/* size of largest swap device */
#define VM_NSWDEV	4
	{ "_nswdev" },	/* number of swap devices */
#define VM_DMMAX	5
	{ "_dmmax" },	/* maximum size of a swap block */
#define	V_MOUNTLIST	6
	{ "_mountlist" },	/* address of head of mount list. */
#define V_NUMV		7
	{ "_numvnodes" },
#define	FNL_NFILE	8
	{"_nfiles"},
#define FNL_MAXFILE	9
	{"_maxfiles"},
#define NLMANDATORY FNL_MAXFILE	/* names up to here are mandatory */
#define X_CNT           10
	{ "_cnt" },                /* struct vmmeter cnt */
#define N_IFNET         11
        { "_ifnet" },
#ifndef HAVE_DEVSTAT
#define X_DKXFER        12
        { "_dk_xfer" },
#endif
	{ "" },
};

#define NPROCS 8

#define	SVAR(var) __STRING(var)	/* to force expansion */
#define	KGET(idx, var)							\
	KGET1(idx, &var, sizeof(var), SVAR(var))
#define	KGET1(idx, p, s, msg)						\
	KGET2(nl[idx].n_value, p, s, msg)
#define	KGET2(addr, p, s, msg)						\
	if (kvm_read(kd, (u_long)(addr), p, s) != s)			\
		warnx("cannot read %s: %s", msg, kvm_geterr(kd))


/* Called at the beginning to inquire how many bars are needed. */
int
num_bars()
{
	return pcount;
}

/*
 * Indicates how many levels each bar has.  For most machines, each bar will
 * have the same stuff.  But one can, for instance, display memory use on one
 * bar, processor levels on others, etc.
 */
void
bar_items(nbars, items)
int nbars;
int items[];    /* nbars items in this */
{
    int i;
    for (i=0; i<NPROCS; i++)
      items[i] = 10;
}

/* Called after num_bars to ask for the bar names */
/* ARGSUSED */
char **
label_bars(nbars)
{
	static char *names[NPROCS];

	if (cpuflag) names[cpuflag-1] = "CPU  ";
	if (loadflag) names[loadflag-1] = "LOAD ";
	if (memflag) names[memflag-1] = "MEM  ";
	if (swapflag) names[swapflag-1] = "SWAP ";
	if (diskflag) names[diskflag-1] = "DISK ";
	if (pktflag) names[pktflag-1] = "PKT  ";
	if (nfsflag) {
	  names[nfsflag-1] = "NFS C";
	  names[nfsflag] = "NFS S";
	}

	return names;
}

/* 
 *  Called after the bars are created to perform any machine dependent
 *  initializations.
 */
/* ARGSUSED */
void
init_bars(nbars)
int nbars;
{
	int ch, i, quit, ret;
	char buf[_POSIX2_LINE_MAX];
	
	if ((kd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, buf)) == 0)
		errx(1, "kvm_openfiles: %s", buf);
	if ((ret = kvm_nlist(kd, nl)) != 0) {
		if (ret == -1)
			errx(1, "kvm_nlist: %s", kvm_geterr(kd));
		for (i = quit = 0; i <= NLMANDATORY; i++)
			if (!nl[i].n_value) {
				quit = 1;
				warnx("undefined symbol: %s\n", nl[i].n_name);
			}
		if (quit) exit(1);
	}

        for (i = 0; i < CPUSTATES; i ++) cp_old[i] = 0;
        packets.in_out = packets.collisions = 0;
#ifdef HAVE_DEVSTAT
	init_devstat();
#endif

}

/* 
 *  This procedure gets called every interval to compute and display the
 *  bars. It should call draw_bar() with the bar number, the array of
 *  integer values to display in the bar, and the number of values in
 *  the array.
 */
/* ARGSUSED */
void
display_bars(nbars)
{
	register int i;
	int	states[10], swap_used, swap_avail, pagesize;
	static	int swap_interval= (-1), pageshift;
	double load, tmp;
	struct vmmeter sum;
	int pkt_in_out;
	int total_xfers=0;
	int mib[3], size;

	/* NPROCS=0, CPU */
  if (cpuflag) {
#define delta(cpustate) ((int) (cp_time[(cpustate)] - cp_old[(cpustate)]))
	for (i=0; i<10; i++)
	    states[i] = 0;
	getkval(nl[X_CP_TIME].n_value, (int *)cp_time, sizeof(cp_time), "_cp_time");
	states[1] = delta(CP_SYS);
	states[4] = delta(CP_INTR);
	states[7] = delta(CP_NICE);
	states[8] = delta(CP_USER);
	states[9] = delta(CP_IDLE);
	draw_bar(cpuflag-1, states, 10);
	for (i = 0; i < CPUSTATES; i ++)
		cp_old[i] = cp_time[i];
  }

	/* NPROCS=1, load */
  if (loadflag) {
	for (i=0; i<10; i++)
	    states[i] = 0;
	getloadavg(&load, 1);
	load *= 100;
	scale_bar(1000, (int)load, 100, states, 0);
	draw_bar(loadflag-1, states, 10);
  }
    	
	/* NPROCS=2, Memory */
  if (memflag) {
#define pagetok(size) ((size) << pageshift)
#define LOG1024 10
	for (i=0; i<10; i++)
	    states[i] = 0;
        pagesize = getpagesize();
        pageshift = 0;
        while (pagesize > 1) {
          pageshift++;
          pagesize >>= 1;
	}
	pageshift -= LOG1024;
        getkval(nl[X_CNT].n_value, (int *)(&sum), sizeof(sum), "_cnt");

        /* convert memory stats to Kbytes */
        states[1] = pagetok(sum.v_active_count);
        states[4] = pagetok(sum.v_inactive_count);
        states[6] = pagetok(sum.v_wire_count);
        states[7] = pagetok(sum.v_cache_count);
        states[9] = pagetok(sum.v_free_count);
	draw_bar(memflag-1, states, 10);
  }
  
	/* NPROCS=3, swap */
  if (swapflag) {
	if (swap_interval<0) {
	  for (i=0; i<10; i++)
	    states[i] = 0;
#ifdef USE_PSTAT
	  /* if swap detection code breaks b'cos of new kernel changes,
	     define this to use /usr/sbin/pstat instead, but slower */
	  swapmode_pstat(&swap_used, &swap_avail);
#else
	  swapmode(&swap_used, &swap_avail);
#endif
	  states[2] = swap_used;
	  states[9] = swap_avail-swap_used;
	  draw_bar(swapflag-1, states, 10);
	  swap_interval = 10;
	} else swap_interval--;
  }
    
	/* NPROCS=4, harddisk access */
  if (diskflag) {
	for (i=0; i<10; i++)
	    states[i] = 0;
#ifdef HAVE_DEVSTAT
	total_xfers = get_devstat();
#else
	getkval(nl[X_DKXFER].n_value, (int *) disk_xfer, sizeof(disk_xfer), "_dk_xfer");
	for (i=0; i < DK_NDRIVE; i++) {
	  tmp = disk_xfer[i];
	  disk_xfer[i] -= old_disk_xfer[i];
	  old_disk_xfer[i] = tmp;
	}
	for(i=0; i < DK_NDRIVE; i++)
	  total_xfers += (int) disk_xfer[i];
#endif
	scale_bar(120, total_xfers, 12, states, 0);
	draw_bar(diskflag-1, states, 10);
  }
    
	/* NPROCS=5, internet packet */
  if (pktflag) {
	for (i=0; i<10; i++)
	    states[i] = 0;
	getkval(nl[N_IFNET].n_value, (int *) &ifnetaddr, sizeof(ifnetaddr), "_ifnet");
	old_packets = packets;
        packets.in_out = packets.collisions = 0;
        while (ifnetaddr) {
            kvm_read(kd, ifnetaddr, &ifnet, sizeof ifnet );
            packets.in_out += ((ifnet.if_ipackets + ifnet.if_opackets) >> 2);
            packets.collisions += ifnet.if_collisions;
#if (__FreeBSD_version >= 300000) 
	    ifnetaddr = (u_long) ifnet.if_link.tqe_next; 
#else 
	    ifnetaddr = (u_long) ifnet.if_next; 
#endif
        }
	pkt_in_out = packets.in_out - old_packets.in_out;
	states[0] = packets.collisions - old_packets.collisions;
	scale_bar(600, pkt_in_out, 65, states, 1);
	draw_bar(pktflag-1, states, 10);
  }
  
	/* NPROCS=6,7 NFS client/server */
  if (nfsflag) {
	for (i=0; i<10; i++)
	    states[i] = 0;
	size = sizeof(nfsstats);
#if (__FreeBSD_version < 199512)
	mib[0] = CTL_FS;
#else
	mib[0] = CTL_VFS;
#endif
#if (__FreeBSD_version >= 300003) /* ?? */
	mib[1] = MNT_EXPORTED;
#else
	mib[1] = MNT_NFS;
#endif
	mib[2] = NFS_NFSSTATS;
	if (sysctl( mib, 3, &nfsstats, &size, NULL, 0) < 0)
	    goto nfs_out;
	else {
	    old_nfsStats = nfsStats;

	    nfsStats.nfsClient = nfsstats.rpccnt[NFSPROC_GETATTR] + 
	                         nfsstats.rpccnt[NFSPROC_SETATTR] +
	                         nfsstats.rpccnt[NFSPROC_LOOKUP] +
	                         nfsstats.rpccnt[NFSPROC_READLINK] +
		                 nfsstats.rpccnt[NFSPROC_READ] +
	                         nfsstats.rpccnt[NFSPROC_WRITE] +
	                         nfsstats.rpccnt[NFSPROC_CREATE] +
	                         nfsstats.rpccnt[NFSPROC_REMOVE] +
	                         nfsstats.rpccnt[NFSPROC_RENAME] +
	                         nfsstats.rpccnt[NFSPROC_LINK] +
	                         nfsstats.rpccnt[NFSPROC_SYMLINK] +
	                         nfsstats.rpccnt[NFSPROC_MKDIR] +
	                         nfsstats.rpccnt[NFSPROC_RMDIR] + 
#if (__FreeBSD_version <= 199504)
				 nfsstats.rpccnt[NFSPROC_READDIR] +
				 nfsstats.rpccnt[NFSPROC_STATFS] +
	                         nfsstats.rpccnt[NQNFSPROC_READDIRLOOK] +
	                         nfsstats.rpccnt[NQNFSPROC_GETLEASE] +
	                         nfsstats.rpccnt[NQNFSPROC_VACATED] +
	                         nfsstats.rpccnt[NQNFSPROC_EVICTED];
#else
				 nfsstats.rpccnt[NFSPROC_READDIR];
#endif
	    nfsStats.nfsServer = nfsstats.srvrpccnt[NFSPROC_GETATTR] + 
	                         nfsstats.srvrpccnt[NFSPROC_SETATTR] +
	                         nfsstats.srvrpccnt[NFSPROC_LOOKUP] +
	                         nfsstats.srvrpccnt[NFSPROC_READLINK] +
	                         nfsstats.srvrpccnt[NFSPROC_READ] +
	                         nfsstats.srvrpccnt[NFSPROC_WRITE] +
	                         nfsstats.srvrpccnt[NFSPROC_CREATE] +
	                         nfsstats.srvrpccnt[NFSPROC_REMOVE] +
	                         nfsstats.srvrpccnt[NFSPROC_RENAME] +
	                         nfsstats.srvrpccnt[NFSPROC_LINK] +
	                         nfsstats.srvrpccnt[NFSPROC_SYMLINK] +
	                         nfsstats.srvrpccnt[NFSPROC_MKDIR] +
	                         nfsstats.srvrpccnt[NFSPROC_RMDIR] + 
#if (__FreeBSD_version <= 199504)
				 nfsstats.srvrpccnt[NFSPROC_READDIR] +
				 nfsstats.srvrpccnt[NFSPROC_STATFS] +
	                         nfsstats.srvrpccnt[NQNFSPROC_READDIRLOOK] +
	                         nfsstats.srvrpccnt[NQNFSPROC_GETLEASE] +
	                         nfsstats.srvrpccnt[NQNFSPROC_VACATED] +
	                         nfsstats.srvrpccnt[NQNFSPROC_EVICTED];
#else
				 nfsstats.srvrpccnt[NFSPROC_READDIR];
#endif
	}
	scale_bar(250, nfsStats.nfsClient-old_nfsStats.nfsClient, 25, states, 0);
	draw_bar(nfsflag-1, states, 10);
	for (i=0; i<10; i++)
	    states[i] = 0;
	scale_bar(250, nfsStats.nfsServer-old_nfsStats.nfsServer, 25, states, 0);
	draw_bar(nfsflag, states, 10);
  }
nfs_out:

}


/* swapmode is derived from freebsd's pstat source ...
 * swapmode is based on a program called swapinfo written
 * by Kevin Lahey <kml@rokkaku.atl.ga.us>.
 */
void
swapmode(int *used, int *avail)
{
	char *header;
	int hlen, nswap, nswdev, dmmax;
	int i, div, nfree, npfree;
	struct swdevt *sw;
	long blocksize, *perdev;
	struct rlist head;
#if (__FreeBSD_version >= 199512)  /* FreeBSD after 2.2-960323-SNAP */
	struct rlisthdr swaplist;
	struct rlist *swapptr;
#else
	struct rlist *swaplist;
#endif
	u_long ptr;

	KGET(VM_NSWAP, nswap);
	KGET(VM_NSWDEV, nswdev);
	KGET(VM_DMMAX, dmmax);
#if (__FreeBSD_version >= 199512)
	KGET1(VM_SWAPLIST, &swaplist, sizeof swaplist, "swaplist");
#else
	KGET(VM_SWAPLIST, swaplist);
#endif
	if ((sw = malloc(nswdev * sizeof(*sw))) == NULL ||
	    (perdev = malloc(nswdev * sizeof(*perdev))) == NULL)
		err(1, "malloc");
	KGET1(VM_SWDEVT, &ptr, sizeof ptr, "swdevt");
	KGET2(ptr, sw, nswdev * sizeof(*sw), "*swdevt");

	/* Count up swap space. */
	nfree = 0;
	memset(perdev, 0, nswdev * sizeof(*perdev));
#if (__FreeBSD_version >= 199512)
	swapptr = swaplist.rlh_list;
	while (swapptr) {
#else
	while (swaplist) {
#endif
		int	top, bottom, next_block;

#if (__FreeBSD_version >= 199512)
                KGET2(swapptr, &head, sizeof(struct rlist), "swapptr");
#else
		KGET2(swaplist, &head, sizeof(struct rlist), "swaplist");
#endif
		top = head.rl_end;
		bottom = head.rl_start;

		nfree += top - bottom + 1;

		/*
		 * Swap space is split up among the configured disks.
		 *
		 * For interleaved swap devices, the first dmmax blocks
		 * of swap space some from the first disk, the next dmmax
		 * blocks from the next, and so on up to nswap blocks.
		 *
		 * The list of free space joins adjacent free blocks,
		 * ignoring device boundries.  If we want to keep track
		 * of this information per device, we'll just have to
		 * extract it ourselves.
		 */
		while (top / dmmax != bottom / dmmax) {
			next_block = ((bottom + dmmax) / dmmax);
			perdev[(bottom / dmmax) % nswdev] +=
				next_block * dmmax - bottom;
			bottom = next_block * dmmax;
		}
		perdev[(bottom / dmmax) % nswdev] +=
			top - bottom + 1;

#if (__FreeBSD_version >= 199512)
		swapptr = head.rl_next;
#else
		swaplist = head.rl_next;
#endif
	}

	header = getbsize(&hlen, &blocksize);
	div = blocksize / 512;
	*avail = npfree = 0;
	for (i = 0; i < nswdev; i++) {
		int xsize, xfree;

		/*
		 * Don't report statistics for partitions which have not
		 * yet been activated via swapon(8).
		 */
		if (!(sw[i].sw_flags & SW_FREED))
			continue;

		/* The first dmmax is never allocated to avoid trashing of
		 * disklabels
		 */
		xsize = sw[i].sw_nblks - dmmax;
		xfree = perdev[i];
		*used = xsize - xfree;
		npfree++;
		*avail += xsize;
		continue;
	}

	/*
	 * If only one partition has been set up via swapon(8), we don't
	 * need to bother with totals.
	 */
	*used = *avail - nfree;
	free(sw);
	free(perdev);
}

/*
 *  getkval(offset, ptr, size, refstr) - get a value out of the kernel.
 *	"offset" is the byte offset into the kernel for the desired value,
 *  	"ptr" points to a buffer into which the value is retrieved,
 *  	"size" is the size of the buffer (and the object to retrieve),
 *  	"refstr" is a reference string used when printing error meessages,
 *	    if "refstr" starts with a '!', then a failure on read will not
 *  	    be fatal (this may seem like a silly way to do things, but I
 *  	    really didn't want the overhead of another argument).
 *  	
 */

int getkval(offset, ptr, size, refstr)
unsigned long offset;
int *ptr;
int size;
char *refstr;

{
    if (kvm_read(kd, offset, (char *) ptr, size) != size)
    {
	if (*refstr == '!')
	    return(0);
	else
	    exit(23);
    }
    return(1);
}

void scale_bar(int max, int value, int unit, int *states, int buttom)
{
	int i;

        if (value < max) {
          states[8] = value % unit;
          states[9] = unit - states[8];
          for (i=0; i<(value/unit - buttom); i++)
             states[7-i] = unit;
        } else {
          for (i=1; i<10; i++) states[i] = 10;
        }
}

#ifdef USE_PSTAT
void swapmode_pstat(int *used, int *total)
{
 FILE *pipe;
 char buf[80], *used_str, *total_str;
 if ((pipe = popen("/usr/sbin/pstat -ks", "r")) == NULL) {
    *used = *total = 1;
    return;
 }
 fgets(buf, sizeof(buf), pipe);
 fgets(buf, sizeof(buf), pipe);
 fgets(buf, sizeof(buf), pipe);
 fgets(buf, sizeof(buf), pipe);
 strtok(buf, " ");
 total_str = strtok(NULL, " ");
 used_str = strtok(NULL, " ");
 pclose(pipe);
 *used = atoi(used_str);
 *total = atoi(total_str); 
}
#endif

#ifdef HAVE_DEVSTAT
/* routines which use libdevstat */
/* this is partly taken from FreeBSD - /usr/src/usr.sbin/iostat */
void
init_devstat(void)
{
	/*
	 * Make sure that the userland devstat version matches the kernel
	 * devstat version.
	 */
	if (checkversion() < 0) {
		nodisk++;
		return;
	}

	/* find out how many devices we have */
	if ((num_devices = getnumdevs()) < 0) {
		nodisk++;
		return;
	}

	cur.dinfo = (struct devinfo *)malloc(sizeof(struct devinfo));
	last.dinfo = (struct devinfo *)malloc(sizeof(struct devinfo));
	bzero(cur.dinfo, sizeof(struct devinfo));
	bzero(last.dinfo, sizeof(struct devinfo));

	/*
	 * Grab all the devices.  We don't look to see if the list has
	 * changed here, since it almost certainly has.  We only look for
	 * errors.
	 */
	if (getdevs(&cur) == -1) {
		nodisk++;
		return;
	}

	num_devices = cur.dinfo->numdevs;
	generation = cur.dinfo->generation;

	dev_select = NULL;

	/* only interested in disks */
	matches = NULL;
	if (buildmatch("da", &matches, &num_matches) != 0) {
		nodisk++;
		return;
	}

	if (num_matches == 0)
		select_mode = DS_SELECT_ADD;
	else
		select_mode = DS_SELECT_ONLY;

	/*
	 * At this point, selectdevs will almost surely indicate that the
	 * device list has changed, so we don't look for return values of 0
	 * or 1.  If we get back -1, though, there is an error.
	 */
	if (selectdevs(&dev_select, &num_selected,
		       &num_selections, &select_generation,
		       generation, cur.dinfo->devices, num_devices,
		       matches, num_matches,
		       NULL, 0,
		       select_mode, 10, 0) == -1)
		nodisk++;
}

int
get_devstat(void)
{
	register int dn;
	long double busy_seconds;
	u_int64_t total_transfers;
	struct devinfo *tmp_dinfo;
	int total_xfers = 0;

	if (nodisk == 0) {
		/*
		 * Here what we want to do is refresh our device stats.
		 * getdevs() returns 1 when the device list has changed.
		 * If the device list has changed, we want to go through
		 * the selection process again, in case a device that we
		 * were previously displaying has gone away.
		 */
		switch (getdevs(&cur)) {
		case -1:
			return (0);
		case 1: {
			int retval;

			num_devices = cur.dinfo->numdevs;
			generation = cur.dinfo->generation;
			retval = selectdevs(&dev_select, &num_selected,
					    &num_selections, &select_generation,
					    generation, cur.dinfo->devices,
					    num_devices, matches, num_matches,
					    NULL, 0,
					    select_mode, 10, 0);
			switch(retval) {
			case -1:
				return (0);
			case 1:
				break;
			default:
				break;
			}
			break;
		}
		default:
			break;
		}

		/*
		 * Calculate elapsed time up front, since it's the same for all
		 * devices.
		 */
		busy_seconds = compute_etime(cur.busy_time, last.busy_time);

		/* this is the first time thru so just copy cur to last */
		if (last.dinfo->numdevs == 0) {
			tmp_dinfo = last.dinfo;
			last.dinfo = cur.dinfo;
			cur.dinfo = tmp_dinfo;
			last.busy_time = cur.busy_time;
			return (0);
		}


		for (dn = 0; dn < num_devices; dn++) {
			int di;

			if ((dev_select[dn].selected == 0)
			 || (dev_select[dn].selected > 10))
				continue;

			di = dev_select[dn].position;

			if (compute_stats(&cur.dinfo->devices[di],
				  &last.dinfo->devices[di], busy_seconds,
				  NULL, &total_transfers,
				  NULL, NULL,
				  NULL, NULL, 
				  NULL, NULL)!= 0)
				  break;
			total_xfers += (int)total_transfers;
		}

		tmp_dinfo = last.dinfo;
		last.dinfo = cur.dinfo;
		cur.dinfo = tmp_dinfo;

		last.busy_time = cur.busy_time;

	} else
		/* no disks found ? */
		total_xfers = 0;

	return (total_xfers);
}
#endif /* HAVE_DEVSTAT */
